//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalOalUltility.cc	--- various utility function used in the CAL/OAL
//
//	DESCRIPTION:
//		This file contains various utility function used by the CAL/OAL
//
//	CHANGE HISTORY:
//		UL, 1997-11-04: 1.0.0: - skeleton
//		UL, 1998-08-17: 1.1.0: - closed a tiny memory leak in stringColumnRow
//							   - added stringColumnRow with string argument
//		UL, 1999-11-10: 1.2.0: - embedded everything in namespace
//
//	AUTHOR:
//		U. Lammers, Tue Nov  4 13:57:27 WET 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /data/xmm02/cvsadmin/soc_repos/sas/packages/caloalutils/src/CalOalUtility.h,v 1.9 2001/07/04 13:51:20 ulammers Exp $
//

#ifndef CALOAL_UTILITY_H
#define CALOAL_UTILITY_H

#include <cassert>
#include <cmath>
#include <vector>
#include <cstring>
#include <string>

#include "Tracer.h"
#include "f90cxx.h"
#include "stringutil.h"
#include "errstr.h"
#include "Dal.h"

namespace CalOalUtility {

template<class T> T SQR(T x) { return x*x; }
template<class T> T CUB(T x) { return x*x*x; }

template<class RealType>
int closestInt(RealType v)
{
	PROC
	RETURN static_cast<int>(v < 0. ? std::ceil(v - .5) : std::floor(v + .5));
} // closestInt


inline std::string f90string(const char *s, unsigned len)
{
	PROC
	//	chop-off trailing blanks
	RETURN trim(std::string(s, len));
}


inline char *stringf90(const std::string s, char *f90s, unsigned f90sLen)
{
	PROC
	assert(f90s);

	std::string str(f90sLen, ' ');
	str.replace(0, s.length()-1, s);
	strncpy(f90s, str.c_str(), f90sLen);

	RETURN f90s;
} // stringf90



typedef std::vector<unsigned> UVec;	// without this dummy declaration
									// compilation of function header below
									// breaks (at least on Solaris/g++ 2.7.2

template<class Type, class DopeT>
inline static void setupDope(DopeT &dope, const Type *vec,
							 std::vector<unsigned> &ext)
{
	PROC
	dope.addr			= (Pointer)vec;		// address to first element
	for (unsigned i=0; i<ext.size(); ++i) {
		dope.dim[i].extent	= ext[i];		// extent of that dimension
		dope.dim[i].mult	= 1;			// stride; distance between
											// successive elements
		dope.dim[i].lower	= 0;			// lower bound

		for (unsigned j=0; j<i; ++j)
			dope.dim[j].mult *= ext[i];
	}

	dope.offset = 0;
	for (unsigned i=0; i<ext.size(); ++i) {
		dope.dim[i].mult *= sizeof(Type);
		dope.offset += dope.dim[i].mult * dope.dim[i].lower;
	}
	dope.offset = -dope.offset;
	ENDPROC
} // setupDope()


template<class FromElementType, class ToElementType>
inline void vector2cvec(const std::vector<FromElementType> &v,
						ToElementType *vec, unsigned n)
{
	PROC
	for (unsigned i=0; i<n; ++i)
		vec[i] = static_cast<ToElementType>(v[i]);
	ENDPROC
} // vector2cvec

std::string stringColumnRow(const ColumnData *cd, unsigned long row);
	//	this function used to be inlined but it leads to an error
	//	in the CAL - compiler bug?
	//


void stringColumnRowSet(const ColumnData *cd, unsigned long row,
					    const std::string &str);

template<class T>
T optionalAttribute(const Block *block, const std::string &attName,
					T defaultValue);

template<class T>
std::vector<T> &fillVector(const Table *table, const std::string &colName,
						   std::vector<T> &v);

std::vector<std::string> &fillVector(const Table *table,
									 const std::string &colName,
									 std::vector<std::string> &v);

template<class T>
std::vector<T> &fillVector(const Table *table, const std::string &colName,
						   unsigned row, std::vector<T> &v);

template<class T>
std::vector<vector<T> > &fillMatrix(const Table *table,
									const std::string &colName,
									std::vector<std::vector<T> > &v);

}

#endif

