//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Celestial.h --- matrix/vector operations in the celestial reference frame
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1997-11-13: 0.1.0: - skeleton
//		UL, 1998-??-??: 1.0.0: - working version
//		UL, 1999-02-10: 1.1.0: - new method: AttitudeMatrix::toEuler321/313
//		UL, 1999-07-01: 1.2.0: - added construction of AttitudeMatrix from
//								 body frame unit vectors
//							   - added unary +/- operator for CartesianVectors
//		JBa,1999-07-29: 1.2.1: - added:
//									* EulerType Euler123 & methods/constructors
//									   with it
//									* construction from "Euler axis and angle"
//									  parameterization
//		UL, 1999-09-17: 1.2.2: - added constant M_SQRT_2
//							   - added SQR function
//		UL, 1999-11-08: 1.2.3: - made EulerAngles a class
//		UL, 2000-03-21: 1.3.0: - added EquatorialDirection constructor
//		UL, 2001-03-02: 1.4.0: - added RightAscension/Declination classes
//		UL, 2001-05-11: 1.5.0: - added WCS classes
//		UL, 2001-06-04: 1.6.0: - added
//									EquatorialDirection::toEpoch
//									EquatorialDirection::toApparentDirection
//
//	AUTHOR:
//		U. Lammers, Thu Nov 13 10:42:10 WET 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /data/xmm02/cvsadmin/soc_repos/sas/packages/caloalutils/src/Celestial.h,v 1.12 2001/07/04 13:51:20 ulammers Exp $
//

#ifndef CELESTIAL_H
#define CELESTIAL_H

#include <cassert>
#include <cmath>
#include <iostream>
#include <string>

#include "Tracer.h"

//
//	---( constants )----------------------------------------------------------
//
#ifdef M_PI
	#undef M_PI
#endif
const double M_PI			= 3.14159265358979323846;
const double M_HALF_PI		= .5*M_PI;
const double M_TWO_PI		= 2.*M_PI;
const double M_SQRT_2		= 1.4142135623730950488;	// sqrt(2)
const double M_SQRT_TWO_PI	= 2.5066282746310005024;	// sqrt(2*Pi)
const double M_2SQ2LN2		= 2.3548200450309493;		// 2.*sqrt(2.*ln(2))
const double EPSILON		= 1.e-12;
const double DEG_TO_RAD		= M_PI/180.;
const double RAD_TO_DEG		= 180./M_PI;

//
//	--( EulerAngles )--------------------------------------------------------
//
class AttitudeMatrix;
class EulerAngles {
public:
	enum Type { Euler123, Euler313, Euler321 };
	EulerAngles(double phi = 0, double theta = 0., double psi = 0.,
				Type type = Euler313) : _phi(phi), _theta(theta), _psi(psi),
									   _type(type) { PROC ENDPROC };

	double phi() const { PROC RETURN _phi; }
	void phi(double val) { PROC _phi = val; ENDPROC }
	double theta() const { PROC RETURN _theta; }
	void theta(double val) { PROC _theta = val; ENDPROC }
	double psi() const { PROC RETURN _psi; }
	void psi(double val) { PROC _psi = val; ENDPROC }
	Type type() const { PROC RETURN _type; }
	void type(Type val) { PROC _type = val; ENDPROC }

	AttitudeMatrix toMatrix() const;

	~EulerAngles() { PROC ENDPROC }

private:
	double	_phi, _theta, _psi;
	Type	_type;
};

std::ostream &operator<<(std::ostream &os, const EulerAngles &eulerAngles);


//
//	---( CartesianVector )---------------------------------------------------
//
class CartesianVector {
public:
	CartesianVector(double x=0., double y=0., double z=0.);
	~CartesianVector() { PROC ENDPROC }

	double x() const		{ PROC RETURN _v[0]; }
	double y() const		{ PROC RETURN _v[1]; }
	double z() const		{ PROC RETURN _v[2]; }
	double *data()			{ PROC RETURN _v; }
	double norm() const		{ PROC RETURN std::sqrt(_v[0]*_v[0] + _v[1]*_v[1] +
										  _v[2]*_v[2]); }
	void normalize()		{ PROC double n = norm(); assert(n!=0.);
							  _v[0] /= n; _v[1] /= n; _v[2] /= n; ENDPROC }


	//	vector *= scalar
	CartesianVector	&operator*=(double);

	//	vector /= scalar
	CartesianVector	&operator/=(double);

	//	vector += vector
	CartesianVector	&operator+=(const CartesianVector &);

	//	vector -= vector
	CartesianVector	&operator-=(const CartesianVector &);

	//	scalar product
	friend double operator*(const CartesianVector &, const CartesianVector &);

	//	cross product
	friend CartesianVector crossProd(const CartesianVector &,
									 const CartesianVector &);

private:
	double _v[3];
};

//	unary -/+
CartesianVector operator+(const CartesianVector &);
CartesianVector operator-(const CartesianVector &);

//	binary -/+
CartesianVector operator+(const CartesianVector &,
						  const CartesianVector &);
CartesianVector operator-(const CartesianVector &,
						  const CartesianVector &);

CartesianVector operator*(const CartesianVector &, double);
CartesianVector operator*(double, const CartesianVector &);

//	scalar/cross product
double operator*(const CartesianVector &, const CartesianVector &);
CartesianVector crossProd(const CartesianVector &,
						  const CartesianVector &);

//	input/output
std::ostream &operator<<(std::ostream &, const CartesianVector &);


//
//	---( AttitudeMatrix )----------------------------------------------------
//
class AttitudeMatrix {
public:
	//	construct from matrix elements
	AttitudeMatrix(double a00, double a01, double a02,
				   double a10, double a11, double a12,
				   double a20, double a21, double a22);

	//	construct from unit vectors of body frame in reference frame
	AttitudeMatrix(const CartesianVector &plusX,
				   const CartesianVector &plusY,
				   const CartesianVector &plusZ);

	//	construct from "Euler axis and angle" parameterization
	AttitudeMatrix(const CartesianVector &axis, double angle);

	//	construction from Euler angles
	AttitudeMatrix(double phi=0., double theta=0., double psi=0.,
				   EulerAngles::Type eulerType = EulerAngles::Euler313);
	AttitudeMatrix(const EulerAngles &eulerAngles);

	~AttitudeMatrix() {}

	//	inverse: for orthognal matrix equal to transpose
	AttitudeMatrix inverse() const;

	//	calculate Euler angles
	EulerAngles		toEulerAngles(EulerAngles::Type eulerType
												= EulerAngles::Euler313);

	typedef double Matrix[3][3];
	const Matrix *data() const { PROC RETURN &_matrix; }
	const double *dataFlat() const { PROC RETURN _matrix[0]; }
	const unsigned size() const { PROC RETURN 3*3; }
	const unsigned memSize() const { PROC RETURN size() * sizeof(double); }

	//	Matrix x Vector
	friend CartesianVector operator*(const AttitudeMatrix&,
									 const CartesianVector&);
	friend AttitudeMatrix operator*(const AttitudeMatrix& a,
									const AttitudeMatrix &b);

private:
	EulerAngles	_euler;
	bool		_eulerValid;
	double		_matrix[3][3];

	EulerAngles toEuler123();
	EulerAngles	toEuler313();
	EulerAngles toEuler321();
	void normalize(EulerAngles &eulerAngles) const;
	void init(double phi=0., double theta=0., double psi=0.,
			  EulerAngles::Type eulerType = EulerAngles::Euler313);
};

std::ostream &operator<<(std::ostream &os, const AttitudeMatrix &attMat);

CartesianVector operator*(const AttitudeMatrix &, const CartesianVector &);
AttitudeMatrix operator*(const AttitudeMatrix& a, const AttitudeMatrix &b);


//
//	---( SpacecraftAttitude )------------------------------------------------
//
class SpacecraftAttitude {
public:
	SpacecraftAttitude(double ra = 0., double dec = 0., double apos = 0.)
		: _ra(ra), _dec(dec), _apos(apos) { PROC ENDPROC }

	double ra() const { PROC RETURN _ra; }
	double dec() const { PROC RETURN _dec; }
	double apos() const { PROC RETURN _apos; }

	AttitudeMatrix	toAttitudeMatrix() const;
	EulerAngles		toEulerAngles() const;

	~SpacecraftAttitude() { PROC ENDPROC }

private:
	double	_ra,	//	Right-Acension/Declination of star-tracker viewing
			_dec,	//	direction in mean geocentric equotorial J2000
					//	reference frame; units: radians
			_apos;	//	astronomical position angle: angle between the plane
					//	defined by the viewing direction and the direction to
					//	the celestial North pole and the S/C +X/+Z plane in the
					//	mathematically positive sense around and along the
					//	viewing	direction (e.g.: S/C +Y axis points to North
					//	pole => astronomical position angle is +90 degrees);
					//	unit: radians
};

std::ostream &operator<<(std::ostream &os, const SpacecraftAttitude &att);

//
//	---( RightAscension / Declination )--------------------------------------
//
class RightAscension {
public:
	RightAscension(double ra = 0.) : _ra(ra) {}
	RightAscension(const std::string &raString);
		//
		//	PURPOSE:
		//		Construct from R.A. (J2000) value
		//
		//	ARGUMENTS:
		//		ra			: R.A. value [rad]
		//		raString	: R.A. as string value in the form:
		//						+ real decimal number
		//						+ "HH[:MM[:SS[.SS]]]"


	double angle() const { return _ra; }
		//
		//	RETURN:
		//		R.A. value as angle [rad]


	RightAscension &operator-=(double a) { _ra -= a; return *this; }
	RightAscension &operator+=(double a) { _ra += a; return *this; }
	RightAscension &operator*=(double a) { _ra *= a; return *this; }
	RightAscension &operator/=(double a) { _ra /= a; return *this; }
		//
		//	PURPOSE:
		//		Arithmetic with R.A.
		//


	std::string toString(char sep = '\0') const;
		//
		//	PURPOSE:
		//		Return string representation "xxhxxmxx.xxs" (or fields
		//		separated by 'sep' if not \0) of R.A. value
		//

private:
	double	_ra;
};


inline std::ostream &operator<<(std::ostream &os, const RightAscension &ra)
{
	os<<ra.toString();

	return os;
}


class Declination {
public:
	Declination(double dec = 0.) : _dec(dec) {}
	Declination(const std::string &decString);
		//
		//	PURPOSE:
		//		Construct from Dec (J2000) value
		//
		//	ARGUMENTS:
		//		dec			: Dec value [rad]; -Pi/2 <= dec <= Pi/2
		//		decString	: Dec as string value in the form:
		//						+ real decimal number
		//						+ "+/-DD[:MM[:SS[.SS]]]"
		//


	double angle() const { return _dec; }
		//
		//	RETURN:
		//		Dec value as angle [rad]


	Declination &operator-=(double a) { _dec -= a; return *this; }
	Declination &operator+=(double a) { _dec += a; return *this; }
	Declination &operator*=(double a) { _dec *= a; return *this; }
	Declination &operator/=(double a) { _dec /= a; return *this; }
		//
		//	PURPOSE:
		//		Arithmetic with Dec
		//


	std::string toString(char sep = '\0') const;
		//
		//	PURPOSE:
		//		Return string representation "xxoxx'xx.xxs''" (or fields
		//		separated by 'sep' if not \0) of Dec value
		//

private:
	double	_dec;
};


inline std::ostream &operator<<(std::ostream &os, const Declination &dec)
{
	os<<dec.toString();

	return os;
}

//
//	---( EquatorialDirection )------------------------------------------------
//
class EquatorialDirection {
public:
	EquatorialDirection(const RightAscension &ra = RightAscension(),
						const Declination &dec = Declination(),
						double epoch = 2000.)
		: _ra(ra), _dec(dec), _epoch(epoch) {}
		//
		//	PURPOSE:
		//		Construct from given Right Ascension/Declination/epoch
		//		tuple
		//


	EquatorialDirection(const CartesianVector &v, double epoch = 2000.);
		//
		//	PURPOSE:
		//		Construct from given Cartesian unit vector and epoch
		//


	RightAscension ra() const { return _ra; }
	Declination dec() const { return _dec; }
	double epoch() const { return _epoch; }
		//
		//	PURPOSE:
		//		Return Right Ascension/Declination and epoch of this
		//		object
		//


	EquatorialDirection &toEpoch(double epoch);
		//
		//	PURPOSE:
		//		Precess coordinates to new given epoch
		//


	EquatorialDirection	&toApparentDirection(const CartesianVector &r,
											 double sourceDist);
		//
		//	PURPOSE:
		//		Assuming the object represents a geocentric equatorial
		//		direction to an celestial source at distance <sourceDist>
		//		the method computes the apparent direction to the object
		//		as observed from an observer's standpoint a <r> w.r.t
		//		to the center of the Earth.
		//
		//	ARGUMENTS:
		//		r			: geocentric position vector of the observer [km]
		//		sourceDist	: distance of the celestial source to
		//					  the center of the Earth [km]
		//


	CartesianVector toUnitCartesian() const;
		//
		//	PURPOSE:
		//		Return cartesian unit vector corresponding to current
		//		object.
		//		


	~EquatorialDirection() {};
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor(s).
		//

private:
	RightAscension	_ra;
	Declination		_dec;
	double			_epoch;
};


inline std::ostream &operator<<(std::ostream &os,
								const EquatorialDirection &dir)
{
	os<<dir.ra()<<'/'<<dir.dec();
	return os;
}

//
//	---( WCS related classes )-------------------------------------------------
//
namespace WCS {

class ZenithalProjection {
public:
	ZenithalProjection() {}

	typedef pair<double, double>	NativePlaneCoordinates;
	virtual EquatorialDirection toStandardSystem(
										const NativePlaneCoordinates &pos) = 0;

	virtual NativePlaneCoordinates toNativeSystem(
										const EquatorialDirection &dir) = 0;

	virtual ~ZenithalProjection() {}
};

class TangentialGnomonicProjection : public ZenithalProjection {
public:
	TangentialGnomonicProjection(const EquatorialDirection &refPoint,
								 double refX, double deltaX,
								 double refY, double deltaY,
								 double longPole = M_PI);

	EquatorialDirection toStandardSystem(const NativePlaneCoordinates &pos);
	NativePlaneCoordinates toNativeSystem(const EquatorialDirection &dir);

	~TangentialGnomonicProjection() {}

private:
	double	_raRef, _decRef,
			_refX, _deltaX,
			_refY, _deltaY,
			_longPole;

	double	_sinDecRef, _cosDecRef;	//	derived
};

}

#endif
