//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997 - 2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	KeplerOrbitSimulator.h --- simulates objects in Kepler orbit around a body
//
//	DESCRIPTION:
//		This is a simple hack to mimic the behavior of a rigid small body which
//		moves in a Keplerian orbit around a massive second body.
//
//	CHANGE HISTORY:
//		UL, 1999-06-11: 1.0.0: - first version
//		UL, 2001-01-05: 1.1.0: - added eccentricity to orbital elements
//
//	$Header: /data/xmm02/cvsadmin/soc_repos/sas/packages/caloalutils/src/KeplerOrbitSimulator.h,v 1.2 2001/01/09 11:20:39 ulammers Exp $
//

#ifndef ORBITSIMULATOR_H
#define ORBITSUMULATOR_H

#include "Celestial.h"

static const double EARTH_RADIUS = 6378.137;		//	[km]

struct OrbitalElements {
	double	period,				//	s
			apogeeHeight,		//	km
			perigeeHeight,		//	km
			eccentricity,		//	0<e<1
			inclincation,		//	orbit inclination [rad]
			ascendingNode,		//	R.A. of ascending node [rad]
			perigeeArgument;	//	argument of perigee [rad]
};


class KeplerOrbitSimulator {
public:
	KeplerOrbitSimulator(const OrbitalElements &orbitalElements,
						 double initalPerigeePass = 0.);
		//
		//	PURPOSE:
		//		Construct a simulator object which can calculate the
		//		position and velocity of an object which moves along
		//		a Keplerian trajectory around a massive body with
		//		constant angular velocity
		//
		//	ARGUMENTS:
		//		In:
		//			orbitalElements		: structure with quantities defining
		//								  the Kepler orbit
		//			initialPerigeePass	: point in time defining the inital
		//								  pass of the object through the
		//								  perigee such that
		//									KeplerOrbitSimulator::position(
		//										initalPerigeePass +
		//											n * orbitalElements.T)
		//								  gives the position of the perigee
		//


	virtual CartesianVector position(double time);
		//
		//	PURPOSE:
		//		Return the position of the object as a Cartesian vector
		//		in the Earth-centered equatorial J2000 reference system
		//		(units km) at a given point in time
		//
		//	ARGUMENTS:
		//		In:
		//			time	: time for which position is requested [s]
		//


	virtual CartesianVector velocity(double time);
		//
		//	PURPOSE:
		//		Return the velocity of the object as a Cartesian vector
		//		in the Earth-centered equatorial J2000 reference system
		//		(units km/s) at a given point in time
		//
		//	ARGUMENTS:
		//		In:
		//			time	: time for which position is requested [s]
		//


	const OrbitalElements& orbitalElements() const { PROC
													 RETURN _orbitalElements; }
		//
		//	PURPOSE:
		//		Return orbital elements this simulator is valid for
		//


	virtual ~KeplerOrbitSimulator() { PROC ENDPROC }
		//
		//	PURPOSE:
		//		Destroy KeplerOrbitSimulator object constructed with
		//		above constructor
		//

protected:
	KeplerOrbitSimulator(double initalPerigeePass = 0);
		//
		//	PURPOSE:
		//		Define simulator w/o orbital elements; these must
		//		be set explicity with below method before
		//		first usage of position()/velocity()
		//


	void orbitalElements(const OrbitalElements &orbitalElements);
		//
		//	PURPOSE:
		//		Define orbital elements for this object
		//

private:
	OrbitalElements	_orbitalElements;		//	orbital elements
	double			_initialPerigeePass;	//	first perigee passage
	AttitudeMatrix	_a313;					//	rotation matrix
	double			_phi0,					//	phase 0
					_omega,					//	angular velocity
					_a,						//	semi-major axis
					_b,						//	semi-minor axis
					_c;						//	difference center - focus
											//	ellipse
	bool _orbitalElementsSet;
	
	//
	//	methods
	//
	void init();
};

#endif
