//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	StateVariable.h	--- classes which define variables for a class state
//
//	DESCRIPTION:
//		Class delarations for variables decribing the state of an object.
//		A class variable possesses a value which can be set/retrieved at all
//		times. Two main types of state variables are defined:
//			a) RangeStateVariables:
//					these are associated with a continuous validity range;
//					the actual value of the variable must lie within the
//					allowed range; Example: 0. <= energy <= 15000 keV
//			b) ChoiceRangeVariables:
//					the actual value must be among an explicit list of allowed
//					values; Example: switch = false|true
//									 color = RED|YELLOW|BLUE
//
//	CHANGE HISTORY:
//		UL, 1999-03-22: 1.0.0: - first version
//		UL, 1999-04-12: 1.1.0: - in call to warning() method from error
//								 packages: explicitely converyt argument to
//								 string to resolve name conflict with
//								 global warning function in Qt
//		UL, 2000-04-06: 1.2.0: - added StateVariable()::activate()/deactivate()
//							   - added StateVariable()::notify()
//							   - moved to caloalutils package
//		UL, 2000-09-22: 1.3.0: - added uncertainty-field + corresponding
//								 access/modification methods to StateVariable
//
//	AUTHOR:
//		U. Lammers, Tue Mar 23 17:57:26 MET 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /data/xmm02/cvsadmin/soc_repos/sas/packages/caloalutils/src/StateVariable.h,v 1.6 2001/07/04 13:51:20 ulammers Exp $
//

#ifndef STATE_VARIABLE_H
#define STATE_VARIABLE_H

#include <iostream>
#include <vector>
#include <string>
#include "Tracer.h"
#include "errstr.h"

//
//	----( Rangeable )------------------------------------------------------
//
//		Class representing a range a numeric value must lie in; the range
//		can be closed, semi-closed or open - in the latter case
//		the value is actually unconstrained
//
template<class Type>
class Rangeable {
public:
	Rangeable(bool hasMin = false, bool hasMax = false,
			  Type min = 0, Type max = 0) :
			  _hasMin(hasMin), _hasMax(hasMax), _min(min), _max(max)
			{ PROC ENDPROC }
		//
		//	PURPOSE:
		//		Construct a Rangeable object
		//
		//	ARGUMENTS:
		//		In:
		//			hasMin/hasMax: is the a minimal/maximal allowed value
		//			min/max		 : the minimal/maximal value allowed

	Type min() const { RETURN _min; }
	Type max() const { RETURN _max; }
		//
		//	PURPOSE:
		//		Return minimal/maximal value for that Rangeable
		//

	bool hasMin() const { RETURN _hasMin; }
	bool hasMax() const { RETURN _hasMax; }
		//
		//	PURPOSE:
		//		Does this Rangeable has a minimal/maximal value?
		//


	bool isIn(Type value) const {
		//
		//	PURPOSE:
		//		Does a given value lie inside the bounds of the Rangeable?
		//
		//	ARGUMENTS:
		//		In:
		//			value: the value to be checked
		//
		PROC
		if (_hasMin)
			if (_hasMax)
				ALTRETURNV(_min <= value && _max >= value)
			else
				ALTRETURNV(_min <= value)
		else
			if (_hasMax)
				ALTRETURNV(_max >= value)

		RETURN true;
	}									

	~Rangeable() { PROC ENDPROC }

private:
	bool	_hasMin, _hasMax;
	Type	_min, _max;
};


//
//	PURPOSE:
//		print a Rangeable onto an ostream
//
template<class Type>
std::ostream &operator<<(std::ostream &os, const Rangeable<Type> &r)
{
	PROC

	if (r.hasMin())
		os<<r.min()<<" <= ";
	os<<"value";
	if (r.hasMax())
		os<<" <= "<<r.max();
	if (!r.hasMin() && !r.hasMax())
		os<<" unconstrained";

	RETURN os;
} // operator<<


//
//	----( Choosable )----------------------------------------------------
//
//		Class representing a set of discrete allowed values of a variable
//
template<class Type>
class Choosable {
public:
	Choosable() { PROC ENDPROC }
	Choosable(const std::vector<Type> &allowedValues)
		: _allowedValues(allowedValues) { PROC ENDPROC }
		//
		//	PURPOSE:
		//		Construct a Choosable object
		//
		//	ARGUMENTS:
		//		In:
		//			allowedValues: vector with allowed values
		//

	const std::vector<Type> &allowedValues() const {
		PROC RETURN _allowedValues;
	}
		//
		//	PURPOSE:
		//		Return vector with allowed values
		//

	bool isIn(Type value) {
		//
		//	PURPOSE:
		//		Is a given value among the list of allowed values?
		//
		//	ARGUMENTS:
		//		In:
		//			value: the value to check
		//
		PROC
		unsigned nmax = _allowedValues.size();
		for (unsigned n=0; n<nmax; ++n)
			if (value == _allowedValues[n])
				ALTRETURNV(true);
		RETURN false;
	}

	~Choosable() { PROC ENDPROC }

private:
	std::vector<Type> _allowedValues;
};


//
//	PURPOSE:
//		print a Rangeable onto an ostream
//
template<class Type>
std::ostream &operator<<(std::ostream &os, const Choosable<Type> &c)
{
	PROC

	const std::vector<Type> &v = c.allowedValues();
	for (unsigned n=0; n<v.size(); ++n)
		os<<v[n]<<" ";

	RETURN os;
} // operator<<


//
//	----( StateVariable )---------------------------------------------------
//
//	A state variable object has a name, unit, and comes with a short
//	description. There are methods to set and retrieve the actual value
//	of the variable. The type is the argument of the template.
//
template<class Type>
class StateVariable {
public:
	StateVariable(const std::string &name)
			: _active(false), _name(name), _unit(""),
			  _description(""), _uncertain(false) { PROC ENDPROC }
		//
		//	PURPOSE:
		//		Construct an inactive state variable - for inactive
		//		state variables, only the name can be inquired - all
		//		other operation lead to a fatal error condition
		//
		//	ARGUMENTS:
		//		In:
		//			name		: name of the variable
		//

	StateVariable(Type initialValue, const std::string &name,
				  const std::string unit = "",
				  const std::string description = "",
				  bool uncertain = false)
			: _active(true), _value(initialValue), _name(name), _unit(unit),
			  _description(description), _uncertain(uncertain)
		{ PROC ENDPROC }
		//
		//	PURPOSE:
		//		Construct a state variable object with an active status
		//
		//	ARGUMENTS:
		//		In:
		//			initialValue: inital value of the variable
		//			name		: name of the variable
		//			unit		: unit of the variable
		//			description	: a short description what this variable
		//						  represents
		//			uncertain	: is the value uncertain?
		//

	bool isActive() const { PROC RETURN _active; }
	void activate() const { PROC _active = true; ENDPROC }
	void deactivate() const { PROC _active = false; ENDPROC }
		//
		//	PURPOSE:
		//		Return/modify activation status of state variable
		//

	bool isUncertain() const { PROC RETURN _uncertain; }
	void uncertain() { PROC _uncertain = true; }
	void certain() { PROC _uncertain = false; }
		//
		//	PURPOSE:
		//		Return/modify uncertainty status of state variable
		//


	const std::string &name() const { PROC RETURN _name; }
	const std::string &unit() const { PROC checkActive(); RETURN _unit; }
	const std::string &description() const { PROC checkActive();
										RETURN _description; }
		//
		//	PUPROSE:
		//		Return the name, unit, and description associated with
		//		the state variable - the unit and description can
		//		be inquired only for active variables
		//

	virtual void set(Type value, bool uncertain = false) = 0;
		//
		//	PURPOSE:
		//		Change the value of the variable;
		//
		//	ARGUMENTS:
		//		In:
		//			value	 : the new value
		//			uncertain: is the value uncertain?
		//

	typedef void (*NotifyFunctor)(Type changedValue);
	virtual void notify(NotifyFunctor f) { PROC _notifyFuncVec.push_back(f);
										   ENDPROC };
	virtual void notify() { PROC
							unsigned N = _notifyFuncVec.size();
							for (unsigned n=0; n<N;
								(*_notifyFuncVec[n++])(_value));
							ENDPROC }

	virtual Type get() const { PROC checkActive(); RETURN _value; }
		//
		//	PURPOSE:
		//		Return the actual value of the variable

	virtual ~StateVariable() { PROC ENDPROC };
		//
		//	PURPOSE:
		//		Destroy state variable constructed with above constructor
		//

private:
	bool	_active,
			_readOnly;
protected:
	std::vector<NotifyFunctor>	_notifyFuncVec;
	Type	_value;
	void	checkActive() const {
		PROC
		if (!_active) {
			errstr<<"The state variable `"<<_name<<"' is not active - "
					"requested operation cannot be performed\n"
				  <<Sas::fatal("StateVariable::checkActive");
		}
	}

private:
	std::string	_name, _unit, _description;

protected:
	bool	_uncertain;
};


//
//	PURPOSE:
//		print a StateVariable onto an ostream
//
template<class Type>
std::ostream &operator<<(std::ostream &os, const StateVariable<Type> &var)
{
	PROC
	os<<"name       : "<<var.name()<<endl
	  <<"unit       : "<<var.unit()<<endl
	  <<"description: "<<var.description()<<endl
	  <<"value      : "<<var.get();

	RETURN os;
} // operator<<


//
//	---( RangeStateVariable )-------------------------------------------------
//
//	A RangeStateVariable is a StateVariable which is Rangeable
//
template<class Type>
class RangeStateVariable : public StateVariable<Type>, public Rangeable<Type> {
public:
	RangeStateVariable(const std::string &name)
		: StateVariable<Type>(name), Rangeable<Type>() { PROC ENDPROC }

	RangeStateVariable(Type initialValue,
					   const std::string &name, const std::string &unit,
					   const std::string &description,
					   bool hasMin, bool hasMax, Type min = 0, Type max = 0)
		: StateVariable<Type>(initialValue, name, unit, description),
		  Rangeable<Type>(hasMin, hasMax, min, max)
		{ PROC set(initialValue); ENDPROC }
		//
		//	PURPOSE:
		//		Construct a RangeStateVariable
		//
		//	ARGUMENTS:
		//		see constructors of StateVariable/Rangeable
		//

	void set(Type value, bool uncertain = false) {
		//
		//	PURPOSE:
		//		Set the actual value of the variable; if out-of-bounds
		//		a warning condition will be raised and the value will
		//		remain unchanged.
		//
		PROC
		_uncertain = uncertain;
		if (isIn(value)) {
			_value = value;
			notify();
		} else {
			Rangeable<Type> *r = this;
			errstr<<"The value "<<value
				  <<" for variable `"<<name()
				  <<"' is not in the allowed range: "
				  <<*r<<endl
				  <<Sas::warning("RangeStateVariable::set");
		}
		ENDPROC
	}

	~RangeStateVariable() { PROC ENDPROC }
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor
		//
}; // RangeStateVariable


//
//	PURPOSE:
//		print a RangeStateVariable onto an ostream
//
template<class Type>
std::ostream &operator<<(std::ostream &os, const RangeStateVariable<Type> &var)
{
	PROC
	const StateVariable<Type> &vs = var;
	const Rangeable<Type> &vr = var;

	os<<vs<<endl;
 	os<<"constraints: "<<vr<<endl;

	RETURN os;
} // operator<<



//
//	---( ChoiceStateVariable )-------------------------------------------------
//
//	A ChoiceStateVariable is a StateVariable which is Choosable
//
template<class Type>
class ChoiceStateVariable : public StateVariable<Type>, public Choosable<Type>{
public:
	ChoiceStateVariable(const std::string &name)
		: StateVariable<Type>(name), Choosable<Type>() { PROC ENDPROC }
	ChoiceStateVariable(Type initialValue,
					    const std::string &name, const std::string &unit,
					    const std::string &description,
						const std::vector<Type> &allowedValues)
		: StateVariable<Type>(initialValue, name, unit, description),
		  Choosable<Type>(allowedValues)
		{ PROC set(initialValue); ENDPROC }
		//
		//	PURPOSE:
		//		Construct a ChoiceStateVariable
		//
		//	ARGUMENTS:
		//		see constructors of StateVariable/Choosable
		//

	void set(Type value, bool uncertain = false) {
		//
		//	PURPOSE:
		//		Set the actual value of the variable; if out-of-bounds
		//		a warning condition will be raised and the value will
		//		remain unchanged.
		//
		PROC
		checkActive();
		_uncertain = uncertain;
		if (isIn(value)) {
			_value = value;
			notify();
		} else {
			Choosable<Type> *c = this;
			errstr<<"The value "<<value
				  <<" for variable `"<<name()
				  <<"' is not among the list of allowed "
					"values: "
				  <<*c<<endl
				  <<Sas::warning("ChoiceStateVariable::set");
		}
		ENDPROC
	}

	~ChoiceStateVariable() { PROC ENDPROC }
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor
		//
};

//
//	PURPOSE:
//		print a ChoiceStateVariable onto an ostream
//
template<class Type>
std::ostream &operator<<(std::ostream &os,
						 const ChoiceStateVariable<Type> &var)
{
	PROC
	const StateVariable<Type> &vs = var;
	const Choosable<Type> &vr = var;

	os<<vs<<endl;
 	os<<"constraints: "<<vr<<endl;

	RETURN os;
} // operator<<

#endif
